from SpiffWorkflow.bpmn.parser.TaskParser import TaskParser
from SpiffWorkflow.bpmn.specs.BpmnSpecMixin import BpmnSpecMixin
from SpiffWorkflow.specs.Simple import Simple

from bpmn_dmn.dmn import DMNDecisionRunner
from bpmn_dmn.bpmn import BPMNXMLWorkflowRunner

class _BusinessRuleTask(Simple, BpmnSpecMixin):
    """
    Task Spec for a bpmn:businessTask (DMB Decision Reference) node.
    """

    def __init__(self, wf_spec, name, **kwargs):
        super().__init__(wf_spec, name, **kwargs)

        self.decisionRunner = None
        self.res = None
        self.resDict = None

    def _on_complete_hook(self, my_task):
        super(_BusinessRuleTask, self)._on_complete_hook(my_task)

        self.res = self.decisionRunner.decide(**my_task.data)
        self.resDict = self.res.outputAsDict()
        my_task.data.update(self.resDict)
        my_task.workflow.data.update(self.resDict)

class _BusinessRuleTaskParser(TaskParser):
    dmn_debug=None

    def __init__(self, process_parser, spec_class, node):
        super(_BusinessRuleTaskParser, self).__init__(process_parser, spec_class, node)
        self.decisionRef = None
        self.dmnParser = None

    def parse_node(self):
        task = super(_BusinessRuleTaskParser, self).parse_node()
        for attrib in self.node.attrib:
            if attrib.endswith('decisionRef'):
                self.setTaskDecision(task, self.node.attrib[attrib])
                break

        return task

    def setTaskDecision(self, task, decisionRef):
        if '.' not in decisionRef:
            decisionRef += '.dmn'

        self.decisionRef = decisionRef
        task.decisionRunner = DMNDecisionRunner(self.decisionRef, debug=_BusinessRuleTaskParser.dmn_debug)

class BPMNDMNXMLWorkflowRunner(BPMNXMLWorkflowRunner):
    def __init__(self, *args, dmn_debug=None, **kwargs):
        super(BPMNDMNXMLWorkflowRunner, self).__init__(*args, **kwargs)

        # Add business rule task support to execute decision tables
        _BusinessRuleTaskParser.dmn_debug = dmn_debug
        self.addParserSupport('businessRuleTask', _BusinessRuleTaskParser, _BusinessRuleTask)



from collections import OrderedDict


class Decision:
    def __init__(self, id, name):
        self.id = id
        self.name = name

        self.decisionTables = []

class DecisionTable:
    def __init__(self, id, name):
        self.id = id
        self.name = name

        self.inputs = []
        self.outputs = []
        self.rules = []

class Input:
    def __init__(self, id, label, name, typeRef):
        self.id = id
        self.label = label
        self.name = name
        self.typeRef = typeRef

class InputEntry:
    def __init__(self, id, input):
        self.id = id
        self.input = input

        self.description = ''
        self.text = ''
        self.operators = []

class Output:
    def __init__(self, id, label, name, typeRef):
        self.id = id
        self.label = label
        self.name = name
        self.typeRef = typeRef

class OutputEntry:
    def __init__(self, id, output):
        self.id = id
        self.output = output

        self.description = ''
        self.text = ''
        self.parsedValue = None

class Rule:
    def __init__(self, id):
        self.id = id

        self.description = ''
        self.inputEntries = []
        self.outputEntries = []

    def outputAsDict(self):
        out = OrderedDict()
        for outputEntry in self.outputEntries:
            out[outputEntry.output.label] = outputEntry.parsedValue # TODO: label?

        return out



from SpiffWorkflow.bpmn.parser.ValidationException import ValidationException
from SpiffWorkflow.bpmn.parser.task_parsers import ExclusiveGatewayParser, first, xpath_eval

class CamundaExclusiveGatewayParser(ExclusiveGatewayParser):
    def connect_outgoing(self, outgoing_task, outgoing_task_node, sequence_flow_node, is_default):
        try:
            super(CamundaExclusiveGatewayParser, self).connect_outgoing(outgoing_task, outgoing_task_node, sequence_flow_node, is_default)
        except ValidationException as ex:
            if 'Non-default exclusive outgoing sequence flow without condition' not in str(ex):
                raise

            xpath = xpath_eval(sequence_flow_node)
            condition_expression_node = conditionExpression = first(
                xpath('.//bpmn:conditionExpression'))

            #if conditionExpression is not None:
            #    conditionExpression = conditionExpression.text

            for attrib in condition_expression_node.attrib:
                if attrib.endswith('resource') and condition_expression_node.attrib[attrib]:
                    conditionExpression = condition_expression_node.attrib[attrib]
                    break

            cond = self.parser.parse_condition(conditionExpression, outgoing_task, outgoing_task_node, sequence_flow_node, condition_expression_node, self)

            self.task.connect_outgoing_if(cond, outgoing_task, sequence_flow_node.get('id'), sequence_flow_node.get(
                'name', None), self.parser._parse_documentation(sequence_flow_node, task_parser=self))



from bpmn_dmn.dmn.engine.dmn_parser import DMNParser
from bpmn_dmn.dmn.engine.dmn_engine import DMNEngine

class DMNDecisionRunner:
    def __init__(self, path, debug=None):
        self.path = path

        self.dmnParser = DMNParser(self.path)
        self.dmnParser.parse()

        decision = self.dmnParser.decision
        assert len(decision.decisionTables) == 1, 'Exactly one decision table should exist! (%s)' % (len(decision.decisionTables))

        self.dmnEngine = DMNEngine(decision.decisionTables[0], debug=debug)

    def decide(self, *inputArgs, **inputKwargs):
        return self.dmnEngine.decide(*inputArgs, **inputKwargs)



from decimal import Decimal
from ast import literal_eval
from xml.etree import ElementTree
from datetime import datetime

from bpmn_dmn.dmn.engine.model import Decision, DecisionTable, Input, InputEntry, Output, OutputEntry, Rule

class DMNParser:
    DT_FORMAT = '%Y-%m-%dT%H:%M:%S'

    def __init__(self, path):
        tree = ElementTree.parse(path)
        self._root = tree.getroot() # definitions

        self.decision = None

    def parse(self):
        self.decision = DMNParser.__parseDecision(self._root)

    @staticmethod
    def __parseDecision(root):
        decisionElements = list(root)
        if len(decisionElements) == 0:
            raise Exception('No decisions found')

        if len(decisionElements) > 1:
            raise Exception('Multiple decisions found')

        decisionElement = decisionElements[0]
        assert decisionElement.tag.endswith('decision'), 'Element %r is not of type "decision"' % (decisionElement.tag)

        decision = Decision(decisionElement.attrib['id'], decisionElement.attrib.get('name', ''))

        # Parse decision tables
        DMNParser.__parseDecisionTables(decision, decisionElement)

        return decision

    @staticmethod
    def __parseDecisionTables(decision, decisionElement):
        for decisionTableElement in decisionElement:
            assert decisionTableElement.tag.endswith('decisionTable'), 'Element %r is not of type "decisionTable"' % (decisionTableElement.tag)

            decisionTable = DecisionTable(decisionTableElement.attrib['id'], decisionTableElement.attrib.get('name', ''))
            decision.decisionTables.append(decisionTable)

            # parse inputs
            DMNParser.__parseInputsOutputs(decision, decisionTable, decisionTableElement)

    @staticmethod
    def __parseInputsOutputs(decision, decisionTable, decisionTableElement):
        for element in decisionTableElement:
            if element.tag.endswith('input'):
                input = DMNParser.__parseInput(element)
                decisionTable.inputs.append(input)
            elif element.tag.endswith('output'):
                output = DMNParser.__parseOutput(element)
                decisionTable.outputs.append(output)
            elif element.tag.endswith('rule'):
                rule = DMNParser.__parseRule(decision, decisionTable, element)
                decisionTable.rules.append(rule)
            else:
                raise Exception('Unknown type in decision table: %r' % (element.tag))

    @staticmethod
    def __parseInput(inputElement):
        typeRef = None
        for inputExpression in inputElement:
            assert inputExpression.tag.endswith('inputExpression'), 'Element %r is not of type "inputExpression"' % (inputExpression.tag)

            typeRef = inputExpression.attrib.get('typeRef', '')

        input = Input(inputElement.attrib['id'], inputElement.attrib.get('label', ''), inputElement.attrib.get('name', ''), typeRef)
        return input

    @staticmethod
    def __parseOutput(outputElement):
        output = Output(outputElement.attrib['id'], outputElement.attrib.get('label', ''), outputElement.attrib.get('name', ''), outputElement.attrib.get('typeRef', ''))
        return output

    @staticmethod
    def __parseRule(decision, decisionTable, ruleElement):
        rule = Rule(ruleElement.attrib['id'])

        inputIdx = 0
        outputIdx = 0
        for child in ruleElement:
            # Load description
            if child.tag.endswith('description'):
                rule.description = child.text

            # Load input entries
            elif child.tag.endswith('inputEntry'):
                inputEntry = DMNParser.__parseInputOutputElement(decision, decisionTable, child, InputEntry, inputIdx)
                rule.inputEntries.append(inputEntry)
                inputIdx += 1

            # Load output entries
            elif child.tag.endswith('outputEntry'):
                outputEntry = DMNParser.__parseInputOutputElement(decision, decisionTable, child, OutputEntry, outputIdx)
                rule.outputEntries.append(outputEntry)
                outputIdx += 1

        return rule

    @staticmethod
    def __parseInputOutputElement(decision, decisionTable, element, cls, idx):
        inputOrOutput = (decisionTable.inputs if cls == InputEntry else decisionTable.outputs if cls == OutputEntry else None)[idx]
        entry = cls(element.attrib['id'], inputOrOutput)

        for child in element:
            if child.tag.endswith('description'):
                entry.description = child.text
            elif child.tag.endswith('text'):
                entry.text = child.text

                if cls == InputEntry:
                    entry.operators = list(DMNParser.__parseRef(inputOrOutput.typeRef, entry.text))
                elif cls == OutputEntry:
                    operators = list(DMNParser.__parseRef(inputOrOutput.typeRef, entry.text))
                    assert len(operators) <= 1, 'Normally it is impossible to have multiple values as output column! %s: %r' % (inputOrOutput.typeRef, entry.text)
                    entry.parsedValue = operators[0][1]
                else:
                    raise NotImplementedError(cls.__name__)

        return entry

    @staticmethod
    def __parseRef(typeRef, val):
        if val is None:
            return []

        if typeRef == 'string':
            return DMNParser.__parseString(val)

        elif typeRef == 'boolean':
            return DMNParser.__parseBoolean(val)

        elif typeRef == 'integer':
            return DMNParser.__parseInteger(val)

        elif typeRef in ('long', 'double'):
            return DMNParser.__parseLongOrDouble(val)

        elif typeRef == 'date':
            return DMNParser.__parseDate(val)

        else:
            raise NotImplementedError(typeRef)

    @staticmethod
    def __parseString(val):
        not_ = False
        if 'not' in val:
            not_ = True
            val = val.replace('not(', '').replace(')', '')

        return [('!=' if not_ else '==', literal_eval(val))]

    @staticmethod
    def __parseBoolean(val):
        return [('==', {'true': True, 'false': False}.get(val, None))]

    @staticmethod
    def __parseInteger(val):
        if '..' in val:
            intStrFrm, intStrTo = val[1:-1].split('..')
            return [
                ('>=' if val[0] == '[' else '>', int(intStrFrm)),
                ('<=' if val[-1] == ']' else '<', int(intStrTo))
            ]
        elif val.startswith(('<', '>', '=')):
            vals = val.split(' ')
            return [(vals[0], int(vals[1]))]
        else:
            return [('==', int(val))]

    @staticmethod
    def __parseLongOrDouble(val):
        if '..' in val:
            intStrFrm, intStrTo = val[1:-1].split('..')
            return [
                ('>=' if val[0] == '[' else '>', Decimal(intStrFrm)),
                ('<=' if val[-1] == ']' else '<', Decimal(intStrTo))
            ]
        elif val.startswith(('<', '>', '=')):
            vals = val.split(' ')
            return [(vals[0], Decimal(vals[1]))]
        else:
            return [('==', Decimal(val))]

    @staticmethod
    def __parseDate(val):
        if '..' in val:
            # Example: [date and time("2017-11-03T00:00:00")..date and time("2017-11-04T23:59:59")]
            dtStrFrm, dtStrTo = val[1:-1].split('..')
            dtFrm = DMNParser.__parseDateStr(dtStrFrm)
            dtTo = DMNParser.__parseDateStr(dtStrTo)
            return [
                       ('>=' if val[0] == '[' else '>', dtFrm),
                       ('<=' if val[-1] == ']' else '<', dtTo)
                   ]
        elif val.startswith(('>', '<')):
            # Example: > date and time("2017-11-03T00:00:00")
            # Example: < date and time("2017-11-03T00:00:00")
            operator, dtStr = val.split(' ', 1)
            dt = DMNParser.__parseDateStr(dtStr)
            return [(operator, dt)]
        else:
            return [('==', DMNParser.__parseDateStr(val))]

    @staticmethod
    def __parseDateStr(val):
        # Example: date and time("2017-11-03T00:00:00")
        dtStrVal = val.replace('date and time("', '').replace('")', '')
        return datetime.strptime(dtStrVal, DMNParser.DT_FORMAT)



from abc import abstractmethod
from io import BytesIO
import logging
from xml.etree import ElementTree

from SpiffWorkflow import Task
from SpiffWorkflow.bpmn.parser.BpmnParser import BpmnParser, full_tag
from SpiffWorkflow.bpmn.serializer.BpmnSerializer import BpmnSerializer
from SpiffWorkflow.bpmn.serializer.Packager import Packager
from SpiffWorkflow.bpmn.specs import ExclusiveGateway
from SpiffWorkflow.bpmn.workflow import BpmnWorkflow

from bpmn_dmn.bpmn.camunda import CamundaExclusiveGatewayParser

class InMemoryPackager(Packager):
    """
    Creates spiff's wf packages on the fly.
    """

    @classmethod
    def package_in_memory(cls, workflow_name, workflow_files, editor):
        """
        Generates wf packages from workflow diagrams.
        """

        s = BytesIO()
        p = cls(s, workflow_name, meta_data=[], editor=editor)
        p.add_bpmn_files_by_glob(workflow_files)
        p.create_package()
        return s.getvalue()

class BPMNXMLWorkflowRunner:
    def __init__(self, path, workflowProcessID=None, debug=False, **kwargs):
        self.path = path
        self.debug = debug
        self.kwargs = kwargs

        ETRroot = ElementTree.parse(self.path).getroot() # definitions
        self.workflowProcessID = workflowProcessID or BPMNXMLWorkflowRunner.__getWorkflowProcessID(ETRroot)
        self.workflowEditor = BPMNXMLWorkflowRunner.__getWorkflowEditor(ETRroot)

        self.packager = InMemoryPackager
        if self.workflowEditor == 'Camunda Modeler':
            self.addParserSupport('exclusiveGateway', CamundaExclusiveGatewayParser, ExclusiveGateway.ExclusiveGateway)

        self.workflow = None

    def addParserSupport(self, full_tag_name, parserClass, taskClass):
        self.packager.PARSER_CLASS.OVERRIDE_PARSER_CLASSES[full_tag(full_tag_name)] = (parserClass, taskClass)

    @staticmethod
    def __getWorkflowProcessID(ETRroot):
        processElements = []
        for child in ETRroot:
            if child.tag.endswith('process') and child.attrib.get('isExecutable', False):
                processElements.append(child)

        if len(processElements) == 0:
            raise Exception('No executable process tag found')

        if len(processElements) > 1:
            raise Exception('Multiple executable processes tags found')

        return processElements[0].attrib['id']

    @staticmethod
    def __getWorkflowEditor(ETRroot):
        return ETRroot.attrib['exporter']

    def __do_engine_steps(self):
        assert not self.workflow.read_only
        engine_steps = list(
            [t for t in self.workflow.get_tasks(Task.READY) if self.workflow._is_engine_task(t.task_spec)])
        while engine_steps:
            for task in engine_steps:
                task.complete()

            engine_steps = list([t for t in self.workflow.get_tasks(Task.READY) if self.workflow._is_engine_task(t.task_spec)])

    def start(self, **data):
        package = self.packager.package_in_memory(self.workflowProcessID, self.path, self.workflowEditor)
        workflowSpec = BpmnSerializer().deserialize_workflow_spec(package)
        self.workflow = BpmnWorkflow(workflowSpec, **self.kwargs)
        self.workflow.debug = self.debug

        # Set input data to first ready task
        self.workflow.get_tasks(Task.READY)[0].set_data(**data)

        # self.workflow.do_engine_steps()

        self.__do_engine_steps()

    def getEndEventName(self):
        endTask = self.workflow.get_tasks()[-1]
        parent = endTask.parent

        while parent and parent.task_spec.name.endswith(('EndJoin', 'End')):
            parent = parent.parent

        if parent:
            return parent.task_spec.name

    def getData(self):
        return self.workflow.data



from distutils.core import setup
from setuptools import setup, find_packages

setup(
  name              = 'bpmn_dmn',
  packages          = find_packages(),
  version           = '0.1.7',
  description       = 'A library to execute BPMN Workflows and DMN Decision Tables',
  author            = 'Denny Weinberg',
  author_email      = 'denny.weinberg@labsolution.lu',
  url               = 'https://github.com/labsolutionlu/bpmn_dmn',
  download_url      = 'https://github.com/labsolutionlu/bpmn_dmn/archive/0.1.5.zip',
  keywords          = ['bpmn', 'dmn', 'camunda'],
  classifiers       = [],
  install_requires  = ['SpiffWorkflow'],
  zip_safe          = True,
)



import logging

# Important for locals (eval)
from decimal import Decimal
import datetime
# / Important for locals (eval)

class DMNEngine:
    def __init__(self, decisionTable, debug=None):
        self.decisionTable = decisionTable
        self.debug = debug

        self.logger = logging.getLogger('DMNEngine')
        if not self.logger.handlers:
            self.logger.addHandler(logging.StreamHandler())
        self.logger.setLevel(getattr(logging, 'DEBUG' if debug else 'INFO'))

    def decide(self, *inputArgs, **inputKwargs):
        for rule in self.decisionTable.rules:
            self.logger.debug('Checking rule %s (%s)...' % (rule.id, rule.description))

            res = self.__checkRule(rule, *inputArgs, **inputKwargs)
            self.logger.debug(' Match? %s' % (res))
            if res:
                self.logger.debug(' Return %s (%s)' % (rule.id, rule.description))
                return rule

    def __checkRule(self, rule, *inputData, **inputKwargs):
        for idx, inputEntry in enumerate(rule.inputEntries):
            input = self.decisionTable.inputs[idx]

            self.logger.debug(' Checking input entry %s (%s: %s)...' % (inputEntry.id, input.label, inputEntry.operators))

            for operator, parsedValue in inputEntry.operators:
                if parsedValue is not None:
                    inputVal = DMNEngine.__getInputVal(inputEntry, idx, *inputData, **inputKwargs)
                    if isinstance(parsedValue, Decimal) and not isinstance(inputVal, Decimal):
                        self.logger.warning('Attention, you are comparing a Decimal with %r' % (type(inputVal)))

                    expression = '%r %s %r' % (inputVal, operator, parsedValue)

                    self.logger.debug(' Evaludation expression: %s' % (expression))
                    if not eval(expression):
                        return False  # Value does not match
                    else:
                        continue  # Check the other operators/columns
                else:
                    # Empty means ignore decision value
                    self.logger.debug(' Value not defined')
                    continue  # Check the other operators/columns

        self.logger.debug(' All inputs checked')
        return True

    @staticmethod
    def __getInputVal(inputEntry, idx, *inputData, **inputKwargs):
        """
        The input of the decision method can be args or kwargs.
        This function tries to extract the input data from args if passed,
         otherwise from kwargs using the label of the decision input column as mapping

        :param inputEntry:
        :param idx:
        :param inputData:
        :param inputKwargs:
        :return:
        """

        return inputData[idx] if inputData else inputKwargs[inputEntry.input.label] # TODO: label?
